@@ -4,25 +4,19 @@ Utility functions for general interactions with Brew and Builds
 from __future__ import absolute_import, print_function, unicode_literals
 
 # stdlib
-import ast
 import time
-import datetime
 import subprocess
-from multiprocessing.dummy import Pool as ThreadPool
-from multiprocessing import cpu_count
 from multiprocessing import Lock
-import shlex
 import traceback
 from typing import List, Tuple, Dict, Optional
 
-from . import exectools
 from . import logutil
 
 # 3rd party
-import click
 import koji
 import koji_cli.lib
 
+
 logger = logutil.getLogger(__name__)
 
 # ============================================================================
@@ -134,3 +128,42 @@ def list_archives_by_builds(build_ids: List[int], build_type: str, session: koji
                 continue
             tasks.append(m.listArchives(buildID=build_id, type=build_type))
     return [task.result if task else None for task in tasks]
+
+
+# Map that records the most recent change for a tag.
+# Maps tag_id to a list containing the most recent change
+# event returned by koji's tagHistory API.
+latest_tag_change_cache = {}
+cache_lock = Lock()
+
+
+def tags_changed_since_build(runtime, koji_client, build, tag_ids):
+    """
+    :param build:  A build information dict returned from koji getBuild
+    :param tag_ids: A list of tag ids (or tag names) which should be assessed
+    :return: If any of the tags have changed since the specified event, returns a
+                list of information about the tags. If no tags have changed since
+                event, returns an empty list
+    """
+    build_nvr = build['nvr']
+    build_event_id = build['creation_event_id']
+
+    result = []
+    for tid in tag_ids:
+        with cache_lock:
+            tag_changes = latest_tag_change_cache.get(tid, None)
+
+        if tag_changes is None:
+            # koji returns in reverse chronological order. So we are retrieving the most recent.
+            tag_changes = koji_client.tagHistory(tag=tid, queryOpts={'limit': 1})
+            with cache_lock:
+                latest_tag_change_cache[tid] = tag_changes
+
+        if tag_changes:
+            tag_change = tag_changes[0]
+            tag_change_event_id = tag_change['create_event']
+            if tag_change_event_id > build_event_id:
+                result.append(tag_change)
+
+    runtime.logger.debug(f'Found that build of {build_nvr} (event={build_event_id}) occurred before tag changes: {result}')
+    return result
