@@ -10,6 +10,7 @@ import io
 import pathlib
 import json
 import logging
+import glob
 
 from . import exectools
 from .pushd import Dir
@@ -19,6 +20,7 @@ from .metadata import Metadata
 from .model import Missing
 from doozerlib.exceptions import DoozerFatalError
 from doozerlib.source_modifications import SourceModifierFactory
+from doozerlib import util
 
 RELEASERS_CONF = """
 [{target}]
@@ -53,6 +55,8 @@ class RPMMetadata(Metadata):
         self.tag = None
         self.build_status = False
         self.pre_init_sha = None
+        # This will be set to True if the source contains embargoed (private) CVE fixes. Defaulting to False for distgit only repos.
+        self.private_fix = False
 
         # If populated, extra variables that will added as os_git_vars
         self.extra_os_git_vars = {}
@@ -60,12 +64,15 @@ class RPMMetadata(Metadata):
         if clone_source:
             self.source_path = self.runtime.resolve_source(self)
             self.source_head = self.runtime.resolve_source_head(self)
-
             with Dir(self.source_path):
                 # gather source repo short sha for audit trail
-                rc, out, _ = exectools.cmd_gather(["git", "rev-parse", "HEAD"])
+                out, _ = exectools.cmd_assert(["git", "rev-parse", "HEAD"])
                 source_full_sha = out.strip()
 
+                # Determine if the source contains private fixes by checking if the private org branch commit exists in the public org
+                if self.public_upstream_branch:
+                    self.private_fix = not util.is_commit_in_public_upstream(source_full_sha, self.public_upstream_branch, self.source_path)
+
                 # If this is a go project, parse the Godeps for points of interest
                 godeps_file = pathlib.Path(self.source_path, 'Godeps', 'Godeps.json')
                 if godeps_file.is_file():
@@ -159,7 +166,7 @@ class RPMMetadata(Metadata):
                 exectools.cmd_assert(f'git reset --hard {self.pre_init_sha}')
                 return
 
-            if not build_spec.push_release_tag:
+            if build_spec.push_release_tag:
                 # If we are configured to push a tag back to the repo, tag the sha we know is there (pre-init)
                 # and push it back.
                 exectools.cmd_assert(["git", "tag", "-am", "Release with doozer", self.tag, self.pre_init_sha])
@@ -424,13 +431,23 @@ class RPMMetadata(Metadata):
         if local:
             raise DoozerFatalError("Local RPM build is not currently supported.")
 
-        with self.runtime.get_dir_lock(self.source_path):
+        # TODO: adjust Doozer's execution for private build
+        if self.private_fix:
+            self.logger.warning("Source contains embargoed fixes.")
+
+        with self.runtime.get_named_lock(self.source_path):
 
             with Dir(self.source_path):
                 # Remember what we were at before tito activity. We may need to revert
                 # to this if we don't push changes back to origin.
                 self.pre_init_sha = exectools.cmd_assert('git rev-parse HEAD')[0].strip()
 
+            if self.public_upstream_branch:
+                if not release.endswith(".p?"):
+                    raise ValueError(f"'release' must end with '.p?' for an image with a public upstream but its actual value is {release}")
+                release = release[:-3]  # strip .p?
+                release += ".p1" if self.private_fix else ".p0"
+
             self.set_nvr(version, release)
             self.tito_setup()
             self.update_spec()
@@ -479,3 +496,32 @@ class RPMMetadata(Metadata):
 
             self.post_build(scratch)
             return self.distgit_key, self.build_status
+
+    def get_package_name(self, default=-1):
+        """
+        Returns the brew package name for the distgit repository. Method requires
+        a local clone.
+        :param default: If specified, this value will be returned if package name could not be found. If
+                        not specified, an exception will be raised instead.
+        :return: The package name if detected in the distgit spec file. Otherwise, the specified default.
+                If default is not passed in, an exception will be raised if the package name can't be found.
+        """
+        specs = glob.glob(f'{self.distgit_repo().distgit_dir}/*.spec')
+        if len(specs) != 1:
+            if default != -1:
+                return default
+            raise IOError('Unable to find .spec file in RPM distgit: ' + self.qualified_name)
+
+        spec_path = specs[0]
+        with open(spec_path, mode='r', encoding='utf-8') as f:
+            for line in f.readlines():
+                if line.lower().startswith('name:'):
+                    return line[5:].strip()  # Exclude "Name:" and then remove whitespace
+
+        if default != -1:
+            return default
+
+        raise IOError(f'Unable to find Name: field in rpm spec: {spec_path}')
+
+    def get_component_name(self, default=-1):
+        return self.get_package_name(default=default)
