@@ -3,6 +3,8 @@ from future import standard_library
 standard_library.install_aliases()
 from future.utils import as_native_str
 from multiprocessing.dummy import Pool as ThreadPool
+from contextlib import contextmanager
+
 import os
 import sys
 import tempfile
@@ -20,19 +22,22 @@ import urllib.parse
 import signal
 import io
 import pathlib
-import subprocess
+import koji
+from typing import Optional
+import time
 
 from doozerlib import gitdata
 from . import logutil
 from . import assertion
 from . import exectools
+from . import dblib
 from .pushd import Dir
 
 from .image import ImageMetadata
 from .rpmcfg import RPMMetadata
 from doozerlib import state
 from .model import Model, Missing
-from multiprocessing import Lock
+from multiprocessing import Lock, RLock
 from .repos import Repos
 from doozerlib.exceptions import DoozerFatalError
 from doozerlib import constants
@@ -110,7 +115,10 @@ def _unpack_tuple_args(func):
 
 class Runtime(object):
     # Use any time it is necessary to synchronize feedback from multiple threads.
-    mutex = Lock()
+    mutex = RLock()
+
+    # Serialize access to the shared koji session
+    koji_lock = RLock()
 
     # Serialize access to the console, and record log
     log_lock = Lock()
@@ -123,9 +131,12 @@ class Runtime(object):
         self.load_disabled = False
         self.data_path = None
         self.data_dir = None
-        self.brew_tag = None
         self.latest_parent_version = False
         self.rhpkg_config = None
+        self._koji_client_session = None
+        self.db = None
+        self.session_pool = {}
+        self.session_pool_available = {}
 
         # Cooperative threads can request exclusive access to directories.
         # This is usually only necessary if two threads want to make modifications
@@ -168,6 +179,10 @@ class Runtime(object):
         # See registry_repo.
         self.source_paths = {}
 
+        # Map of source code repo aliases (e.g. "ose") to a (public_upstream_url, public_upstream_branch) tuple.
+        # See registry_repo.
+        self.public_upstreams = {}
+
         # Map of stream alias to image name.
         self.stream_alias_overrides = {}
 
@@ -207,7 +222,7 @@ class Runtime(object):
         else:
             self.rhpkg_config = ''
 
-    def get_dir_lock(self, absolute_path):
+    def get_named_lock(self, absolute_path):
         with self.dir_locks['']:
             p = pathlib.Path(absolute_path).absolute()  # normalize (e.g. strip trailing /)
             if p in self.dir_locks:
@@ -249,7 +264,8 @@ class Runtime(object):
     def initialize(self, mode='images', clone_distgits=True,
                    validate_content_sets=False,
                    no_group=False, clone_source=True, disabled=None,
-                   config_excludes=None):
+                   config_excludes=None,
+                   upstream_commitish_overrides={}):
 
         if self.initialized:
             return
@@ -294,6 +310,8 @@ class Runtime(object):
 
         self.init_state()
 
+        self.db = dblib.DB(self, self.datastore)
+
         if no_group:
             return  # nothing past here should be run without a group
 
@@ -403,14 +421,13 @@ class Runtime(object):
                         )
                 scanner.matches = regexen
 
-            if self.group_config.brew_tag:
-                # setting this overrides tags made out of branches in specific configs
-                self.brew_tag = self.group_config.brew_tag
-
             exclude_keys = flatten_list(self.exclude)
             image_ex = list(exclude_keys)
             rpm_ex = list(exclude_keys)
             image_keys = flatten_list(self.images)
+
+            self.upstream_commitish_overrides = upstream_commitish_overrides
+
             rpm_keys = flatten_list(self.rpms)
 
             filter_func = None
@@ -463,7 +480,7 @@ class Runtime(object):
 
             if mode in ['images', 'both']:
                 for i in image_data.values():
-                    metadata = ImageMetadata(self, i)
+                    metadata = ImageMetadata(self, i, self.upstream_commitish_overrides.get(i.key))
                     self.image_map[metadata.distgit_key] = metadata
                 if not self.image_map:
                     self.logger.warning("No image metadata directories found for given options within: {}".format(self.group_dir))
@@ -554,6 +571,54 @@ class Runtime(object):
         debug_log_handler.setLevel(logging.DEBUG)
         self.logger.addHandler(debug_log_handler)
 
+    @contextmanager
+    def shared_koji_client_session(self):
+        """
+        Context manager which offers a shared koji client session. You hold a koji specific lock in this context
+        manager giving your thread exclusive access. The lock is reentrant, so don't worry about
+        call a method that acquires the same lock while you hold it.
+        """
+        with self.koji_lock:
+            if self._koji_client_session is None:
+                self._koji_client_session = koji.ClientSession(self.group_config.urls.brewhub)
+            yield self._koji_client_session
+
+    @contextmanager
+    def pooled_koji_client_session(self):
+        """
+        Context manager which offers a koji client session from a limited pool. You hold a lock on this
+        session until you return. It is not recommended to call other methods that acquire their
+        own pooled sessions, because that may lead to deadlock if the pool is exhausted.
+        """
+        session = None
+        session_id = None
+        while True:
+            with self.mutex:
+                if len(self.session_pool_available) == 0:
+                    if len(self.session_pool) < 30:
+                        # pool has not grown to max size;
+                        new_session = koji.ClientSession(self.group_config.urls.brewhub)
+                        session_id = len(self.session_pool)
+                        self.session_pool[session_id] = new_session
+                        session = new_session  # This is what we wil hand to the caller
+                        break
+                    else:
+                        # Caller is just going to have to wait and try again
+                        pass
+                else:
+                    session_id, session = self.session_pool_available.popitem()
+                    break
+
+            time.sleep(5)
+
+        # Arriving here, we have a session to use.
+        try:
+            yield session
+        finally:
+            # Put it back into the pool
+            with self.mutex:
+                self.session_pool_available[session_id] = session
+
     @staticmethod
     def timestamp():
         return datetime.datetime.utcnow().isoformat()
@@ -753,12 +818,44 @@ class Runtime(object):
         if not data_obj:
             raise DoozerFatalError('Unable to resolve image metadata for {}'.format(distgit_name))
 
-        meta = ImageMetadata(self, data_obj)
+        meta = ImageMetadata(self, data_obj, self.upstream_commitish_overrides.get(data_obj.key))
         if add:
             self.image_map[distgit_name] = meta
         return meta
 
+    def resolve_brew_image_url(self, image_name_and_version):
+        """
+        :param image_name_and_version: The image name to resolve. The image can contain a version tag or sha.
+        :return: Returns the pullspec of this image in brew.
+        e.g. "openshift/jenkins:5"  => "registry-proxy.engineering.redhat.com/rh-osbs/openshift-jenkins:5"
+        """
+
+        if self.group_config.urls.brew_image_namespace is not Missing:
+            # if there is a namespace, we need to flatten the image name.
+            # e.g. openshift/image:latest => openshift-image:latest
+            # ref: https://mojo.redhat.com/docs/DOC-1204856
+            url = self.group_config.urls.brew_image_host
+            ns = self.group_config.urls.brew_image_namespace
+            name = image_name_and_version.replace('/', '-')
+            url = "/".join((url, ns, name))
+        else:
+            # If there is no namespace, just add the image name to the brew image host
+            url = "/".join((self.group_config.urls.brew_image_host, image_name_and_version))
+
+        if ':' not in url.split('/')[-1]:
+            # oc image info will return information about all tagged images. So be explicit
+            # in indicating :latest if there is no tag.
+            url += ':latest'
+
+        return url
+
     def resolve_stream(self, stream_name):
+        """
+        :param stream_name: The name of the stream to resolve.
+        :return: Resolves and returns the image stream name into its literal value.
+                This is usually a lookup in streams.yaml, but can also be overridden on the command line. If
+                the stream_name cannot be resolved, an exception is thrown.
+        """
 
         # If the stream has an override from the command line, return it.
         if stream_name in self.stream_alias_overrides:
@@ -769,7 +866,7 @@ class Runtime(object):
 
         return self.streams[stream_name]
 
-    def get_public_upstream(self, remote_git):
+    def get_public_upstream(self, remote_git: str) -> (str, Optional[str]):
         """
         Some upstream repo are private in order to allow CVE workflows. While we
         may want to build from a private upstream, we don't necessarily want to confuse
@@ -780,7 +877,9 @@ class Runtime(object):
         into a new string with the public prefix in its place. If there is not
         applicable mapping, the incoming url will still be normalized into https.
         :param remote_git: The URL to analyze for private repo patterns.
-        :return: An https normalized remote address with private repo information replaced.
+        :return: tuple (url, branch)
+            - url: An https normalized remote address with private repo information replaced.
+            - branch: Optional public branch name if the public upstream source use a different branch name from the private upstream.
         """
         remote_https = util.convert_remote_git_to_https(remote_git)
 
@@ -790,7 +889,10 @@ class Runtime(object):
             # map and keep track of the longest matching private remote.
             target_priv_prefix = None
             target_pub_prefix = None
-            for priv, pub in self.group_config.public_upstreams.items():
+            target_pub_branch = None
+            for upstream in self.group_config.public_upstreams:
+                priv = upstream["private"]
+                pub = upstream["public"]
                 # priv can be a full repo, or an organization (e.g. git@github.com:openshift)
                 # It will be treated as a prefix to be replaced
                 https_priv_prefix = util.convert_remote_git_to_https(priv)  # Normalize whatever is specified in group.yaml
@@ -800,24 +902,25 @@ class Runtime(object):
                     if not target_priv_prefix or len(https_priv_prefix) > len(target_pub_prefix):
                         target_priv_prefix = https_priv_prefix
                         target_pub_prefix = https_pub_prefix
+                        target_pub_branch = upstream.get("public_branch")
 
             if target_priv_prefix:
-                return f'{target_pub_prefix}{remote_https[len(target_priv_prefix):]}'
+                return (f'{target_pub_prefix}{remote_https[len(target_priv_prefix):]}', target_pub_branch)
 
-        return remote_https
+        return (remote_https, None)
 
     def git_clone(self, remote_url, target_dir, gitargs=None, set_env=None, timeout=0):
         gitargs = gitargs or []
         set_env = set_env or []
 
         if self.cache_dir:
-            git_cache_dir = os.path.join(self.cache_dir, self.user, 'git')
+            git_cache_dir = os.path.join(self.cache_dir, self.user or "default", 'git')
             util.mkdirs(git_cache_dir)
             normalized_url = util.convert_remote_git_to_https(remote_url)
             # Strip special chars out of normalized url to create a human friendly, but unique filename
             file_friendly_url = normalized_url.split('//')[-1].replace('/', '_')
             repo_dir = os.path.join(git_cache_dir, file_friendly_url)
-            self.logger.info(f'Cache for {remote_url} going to f{repo_dir}')
+            self.logger.info(f'Cache for {remote_url} going to {repo_dir}')
 
             if not os.path.exists(repo_dir):
                 self.logger.info(f'Initializing cache directory for git remote: {remote_url}')
@@ -825,7 +928,7 @@ class Runtime(object):
                 # If the cache directory for this repo does not exist yet, we will create one.
                 # But we must do so carefully to minimize races with any other doozer instance
                 # running on the machine.
-                with self.get_dir_lock(repo_dir):  # also make sure we cooperate with other threads in this process.
+                with self.get_named_lock(repo_dir):  # also make sure we cooperate with other threads in this process.
                     tmp_repo_dir = tempfile.mkdtemp(dir=git_cache_dir)
                     exectools.cmd_assert(f'git init --bare {tmp_repo_dir}')
                     with Dir(tmp_repo_dir):
@@ -896,9 +999,25 @@ class Runtime(object):
         else:
             return None
 
+        def _get_public_upstream_from_cache(alias):
+            if alias in self.public_upstreams:
+                return self.public_upstreams[alias]
+            # public upstream is not found. source is given by --source or --sources?
+            # attempt to determine private upstream URL and branch name
+            out, _ = exectools.cmd_assert(["git", "remote", "get-url", "origin"])
+            url = out.strip()
+            public_upstream_url, public_upstream_branch = self.get_public_upstream(url)
+            if not public_upstream_branch:  # default to the same branch name as private upstream
+                out, _ = exectools.cmd_assert(["git", "symbolic-ref", "--short", "HEAD"])
+                public_upstream_branch = out.strip()
+            self.public_upstreams[alias] = (public_upstream_url, public_upstream_branch)  # cache it! should be atomic and idempotent
+            return public_upstream_url, public_upstream_branch
+
         self.logger.debug("Resolving local source directory for alias {}".format(alias))
         if alias in self.source_paths:
             self.logger.debug("returning previously resolved path for alias {}: {}".format(alias, self.source_paths[alias]))
+            if self.group_config.public_upstreams:
+                meta.public_upstream_url, meta.public_upstream_branch = _get_public_upstream_from_cache(alias)
             return self.source_paths[alias]
 
         # Where the source will land, check early so we know if old or new style
@@ -912,11 +1031,21 @@ class Runtime(object):
 
         self.logger.debug("checking for source directory in source_dir: {}".format(source_dir))
 
-        with self.get_dir_lock(source_dir):
+        with self.get_named_lock(source_dir):
             if alias in self.source_paths:  # we checked before, but check again inside the lock
                 self.logger.debug("returning previously resolved path for alias {}: {}".format(alias, self.source_paths[alias]))
+                if self.group_config.public_upstreams:
+                    meta.public_upstream_url, meta.public_upstream_branch = _get_public_upstream_from_cache(alias)
                 return self.source_paths[alias]
 
+            url = source_details["url"]
+            clone_branch, _ = self.detect_remote_source_branch(source_details)
+            if self.group_config.public_upstreams:
+                meta.public_upstream_url, meta.public_upstream_branch = self.get_public_upstream(url)
+                if not meta.public_upstream_branch:  # default to the same branch name as private upstream
+                    meta.public_upstream_branch = clone_branch
+                self.public_upstreams[alias] = (meta.public_upstream_url, meta.public_upstream_branch)  # save to cache
+
             # If this source has already been extracted for this working directory
             if os.path.isdir(source_dir):
                 # Store so that the next attempt to resolve the source hits the map
@@ -924,21 +1053,33 @@ class Runtime(object):
                 self.logger.info("Source '{}' already exists in (skipping clone): {}".format(alias, source_dir))
                 return source_dir
 
-            clone_branch, _ = self.detect_remote_source_branch(source_details)
-
-            url = source_details["url"]
-
             self.logger.info("Attempting to checkout source '%s' branch %s in: %s" % (url, clone_branch, source_dir))
             try:
                 self.logger.info("Attempting to checkout source '%s' branch %s in: %s" % (url, clone_branch, source_dir))
                 # clone all branches as we must sometimes reference master /OWNERS for maintainer information
                 gitargs = [
-                    '--no-single-branch', '--branch', clone_branch, '--depth', '1'
+                    '--no-single-branch', '--branch', clone_branch
                 ]
+                if not self.group_config.public_upstreams:
+                    # Do a shallow clone only if public upstreams are not specified
+                    gitargs.extend(['--depth', '1'])
                 self.git_clone(url, source_dir, gitargs=gitargs, set_env=constants.GIT_NO_PROMPTS)
+                if meta.commitish:
+                    # Commit-ish may not be in the history because of shallow clone. Fetch from upstream if not existing.
+                    self.logger.info(f"Determining if commit-ish {meta.commitish} exists")
+                    cmd = ["git", "-C", source_dir, "branch", "--contains", meta.commitish]
+                    rc, _, _ = exectools.cmd_gather(cmd)
+                    if rc != 0:  # commit-ish does not exist
+                        self.logger.info(f"Fetching commit-ish {meta.commitish} from upstream")
+                        cmd = ["git", "-C", source_dir, "fetch", "origin", "--depth", "1", meta.commitish]
+                        exectools.cmd_assert(cmd)
+                # fetch public upstream source
+                if meta.public_upstream_branch:
+                    util.setup_and_fetch_public_upstream_source(meta.public_upstream_url, meta.public_upstream_branch, source_dir)
 
             except IOError as e:
                 self.logger.info("Unable to checkout branch {}: {}".format(clone_branch, str(e)))
+                shutil.rmtree(source_dir)
                 raise DoozerFatalError("Error checking out target branch of source '%s' in: %s" % (alias, source_dir))
 
             # Store so that the next attempt to resolve the source hits the map
@@ -1097,6 +1238,12 @@ class Runtime(object):
             n_threads=n_threads)
 
     def parallel_exec(self, f, args, n_threads=None):
+        """
+        :param f: A function to invoke for all arguments
+        :param args: A list of argument tuples. Each tuple will be used to invoke the function once.
+        :param n_threads: preferred number of threads to use during the work
+        :return:
+        """
         n_threads = n_threads if n_threads is not None else len(args)
         terminate_event = threading.Event()
         pool = ThreadPool(n_threads)
@@ -1117,10 +1264,16 @@ class Runtime(object):
         pool.join()
         return ret
 
+    def get_default_brew_tag(self):
+        return self.branch
+
+    def get_default_candidate_brew_tag(self):
+        return self.branch + '-candidate' if self.branch else None
+
     def builds_for_group_branch(self):
         # return a dict of all the latest builds for this group, according to
         # the branch's candidate tag in brew. each entry is name => tuple(version, release).
-        tag = self.brew_tag if self.brew_tag else '{}-candidate'.format(self.branch)
+        tag = self.get_default_candidate_brew_tag()
         output, _ = exectools.cmd_assert(
             "brew list-tagged --quiet --latest {}".format(tag),
             retries=3,
@@ -1134,9 +1287,8 @@ class Runtime(object):
         return {n: (v, r) for n, v, r in builds}
 
     def scan_distgit_sources(self):
-        builds = self.builds_for_group_branch()  # query for builds only once
         return self.parallel_exec(
-            lambda meta, _: (meta, meta.distgit_repo(autoclone=False).matches_source_commit(builds)),
+            lambda meta, _: (meta, meta.needs_rebuild()),
             self.image_metas() + self.rpm_metas(),
             n_threads=100,
         ).get()
